
package org.saev.deob;

import java.io.*;
import java.util.*;
import java.util.jar.*;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.saev.deob.util.*;

public class Application {
	private HashMap<String, ClassGen> classMap;
	private Manifest manifest;

	/** constructs an empty Application. */
	public Application() {
		classMap = new HashMap<String, ClassGen>();
		manifest = new Manifest();
	}

	/** constructs an Application from a file, directory, or file.
	* @param path an absolute or relative path pointing to a jar,
	* directory, or file. */
	public Application(String path) {
		classMap = new HashMap<String, ClassGen>();
		path = path.toLowerCase();
		
		if(path.endsWith(".jar")) {
			loadJar(path);
		} else {
			File file = new File(path);
 	
			if(file.isDirectory()) {
				loadDirectory(file);
			} else {
				loadClass(file);
			}
		}
		
		if(manifest == null) {
			manifest = new Manifest();
		}
	}

	/** adds all the class files in a jar to this Application.
	* @param path an absolute or relative path to a jar. */
	public void loadJar(String path) {
		JarFile jar = null;
		
		try {
			jar = new JarFile(path);
		} catch(IOException ioe) {
			System.out.println("could not read jar file \"" + path + "\": " + ioe);
		}
		
		try {
			manifest = jar.getManifest();
		} catch(IOException ioe) { }
		
		if (manifest == null) {
			manifest = new Manifest();
		} else {
			scrubManifest(manifest);
		}
		
		for(Enumeration<JarEntry> je = jar.entries(); je.hasMoreElements(); ) {
			JarEntry entry = je.nextElement();
			
			if(!entry.isDirectory()) {
				String entryName = entry.getName();
				
				try {
					InputStream is = new BufferedInputStream(jar.getInputStream(entry));
					
					if(entryName.endsWith(".class")) {
						JavaClass clazz = new ClassParser(is, entryName).parse();
						classMap.put(clazz.getClassName(), new ClassGen(clazz));
					}
					is.close();
				} catch(ClassFormatException cfe) {
					System.out.println("\"" + entryName + "\" is not a valid class: " + cfe);
				} catch(IOException ioe) {
					System.out.println("could not load class \"" + entryName + "\": " + ioe);
				}
			}
		}
	}
	
	private void scrubManifest(Manifest manifest) {
		
		for(Iterator<String> entryIt = manifest.getEntries().keySet().iterator(); entryIt.hasNext(); ) {
			String entry = entryIt.next();
			
			if(entry.endsWith(".class")) {
				Attributes attr = manifest.getAttributes(entry);
				
				for(Iterator attrIt = attr.keySet().iterator(); attrIt.hasNext(); ) {
					String key = ((Attributes.Name) attrIt.next()).toString().toLowerCase();
					
					if (key.indexOf("-digest-") > 0 || key.endsWith("-digest") ||
							key.equals("digest-algorithms") || key.equals("magic")) {
						attrIt.remove();
					}
				}
				
				if (attr.size() == 0) {
					entryIt.remove();
				}
			}
		}
	}

	/** adds all the class files in a directory (recursive) to this Application.
	* @param path an absolute or relative path to a directory. */
	public void loadDirectory(String path) {
		loadDirectory(new File(path));
	}

	/** adds all the class files in a directory (recursive) to this Application.
	* @param file a file object pointing to a directory. */
	public void loadDirectory(File file) {
		
		if(file.isDirectory()) {
			File[] files = file.listFiles(new FileFilter() {
				
				public boolean accept(File f) {
					
					if(f.getName().endsWith(".class") || f.isDirectory()) {
						return true;
					}
					return false;
				}
			});
			
			for(int i = 0; i < files.length; i++) {
				
				if(files[i].isDirectory()) {
					loadDirectory(files[i]);
				}
				loadClass(files[i]);
			}
		} else {
			throw new RuntimeException("file \"" + file + "\" is not a directory");
		}
	}

	/** adds a class file to this Application.
	* @param path an absolute or relative path to a classfile. */
	public void loadClass(String path) {
		loadClass(new File(path));
	}

	/** adds a class file to this Application.
	* @param file a file object pointing to a classfile. */
	public void loadClass(File file) {
		
		if(file.getName().endsWith(".class")) {
			try {
				JavaClass clazz = new ClassParser(file.getPath()).parse();
				classMap.put(clazz.getClassName(), new ClassGen(clazz));
			} catch(ClassFormatException cfe) {
				System.out.println("\"" + file.getName() + "\" is not a valid class: " + cfe);
			} catch(IOException ioe) {
				System.out.println("could not load class \"" + file.getName() + "\": " + ioe);
			}
		}
	}

	/** saves this Application to a directory.
	* @param path an absolute or relative path to a directory. if the directory
	* does not exist, it will be created. */
	public void dumpDirectory(String path) {
		
		for(Iterator<ClassGen> it = classMap.values().iterator(); it.hasNext(); ) {
			ClassGen cgen = it.next();
			String newPath = path + File.separatorChar +
					cgen.getClassName().replace('.', File.separatorChar) + ".class";
			
			try {
				cgen.getJavaClass().dump(newPath);
			} catch(IOException ioe) {
				System.out.println("could not write class \"" + path + "\": " + ioe);
			}
		}
	}

	/** saves this Application to a jar.
	* @param path an absolute or relative path to a jar. the jar will be created. */
	public void dumpJar(String path) {
		FileOutputStream os;
		
		try {
			os = new FileOutputStream(new File(path));
		} catch(FileNotFoundException fnfe) {
			throw new RuntimeException("could not create file \"" + path + "\": " + fnfe);
		}
		JarOutputStream jos;
		Attributes attr = manifest.getMainAttributes();
		
		try {
			
			if(attr.getValue(Attributes.Name.MANIFEST_VERSION) == null) {
				attr.put(Attributes.Name.MANIFEST_VERSION, "1.0");
			}
			
			if(attr.getValue(Attributes.Name.MAIN_CLASS) == null) {
				String mainClassName = getMainClassName();
				
				if(mainClassName != null) {
					setMainAttribute(mainClassName);
				}
			}
			jos = new JarOutputStream(os, manifest);
			
			for(Iterator<ClassGen> classIt  = classMap.values().iterator(); classIt.hasNext(); ) {
				ClassGen cgen = classIt.next();
				jos.putNextEntry(new JarEntry(cgen.getClassName().replace('.', File.separatorChar) + ".class"));
				jos.write(cgen.getJavaClass().getBytes());
				jos.flush();
			}
			jos.close();
		} catch(IOException ioe) {
			throw new RuntimeException(ioe);
		}
	}

	/** gets the name of the main class for this Application.
	* @return the name of the main class for this Application,
	* or null if more than one class with a 'public static final main(String[])
	* is found. */
	public String getMainClassName() {
		Attributes attr = manifest.getMainAttributes();
		
		if(attr != null) {
			String mainClassName;
			
			if((mainClassName = (String) attr.get(Attributes.Name.MAIN_CLASS)) != null) {
				return mainClassName;
			}
		}
		ClassGen cgen = getMainClass();
		return cgen == null ? null : cgen.getClassName();
	}
	
	public ClassGen getMainClass() {
		ClassGen main = null;
		int n = 0;
		
		for(ClassGen cgen: classMap.values()) {
			Method[] methods = cgen.getMethods();
			
			for(int i = 0; i < methods.length; i++) {
				Method m = methods[i];
				
				if(isMainMethod(m)) {
					main = cgen;
					n++;
					break;
				}
			}
		}
		
		if(n == 1) {
			return main;
		}
		return null;
	}

	/** sets the main class attribute.
	* @param className name of the main class. */
	public void setMainAttribute(String className) {
		
		if (className != null) {
			Attributes attr = manifest.getMainAttributes();
			attr.put(Attributes.Name.MAIN_CLASS, className);
		}
	}

	private boolean isMainMethod(Method m) {
		return m.getName().equals("main") && m.getSignature().equals("([Ljava/lang/String;)V")
		&& m.isPublic() && m.isStatic();
	}

	/** gets a ClassGen object by its name. if the class is not stored
	* in this Application, it will check the classpath for it.
	* @param className the name of the class to get.
	* @return a ClassGen object, or null if it could not be found. */
	public ClassGen getClassGen(String className) {
		ClassGen cgen = classMap.get(className);
		
		if(cgen == null) {
			
			try {
				cgen = new ClassGen(Repository.lookupClass(className));
			} catch(ClassNotFoundException cnfe) { }
		}
		return cgen;
	}

	/** adds a class to this Application. it will replace an existing class
	* with the same name.
	* @param cgen the ClassGen to add to this Application. */
	public void setClassGen(ClassGen cgen) {
		classMap.put(cgen.getClassName(), cgen);
	}

	/** gets a JavaClass object by its class name.
	* @param className name of the class to get.
	* @return a JavaClass object, or null if it could not be found. */
	public JavaClass getJavaClass(String className) {
		return getClassGen(className).getJavaClass();
	}
	/** adds a class to this Application. it will replace an existing class
	* with the same name.
	* @param clazz the ClassGen to add to this Application. */
	public void setJavaClass(JavaClass clazz) {
		classMap.put(clazz.getClassName(), new ClassGen(clazz));
	}

	/** gets a MethodGen object from a MethodID.
	* @param mid a MethodID.
	* @return a MethodGen, or null if it could not be found. */
	public MethodGen getMethod(MethodID mid) {
		return getMethod(mid.getClassName(), mid.getName(), mid.getSignature());
	}

	/** gets a MethodGen object from a MethodID.
	* @param className the class the method belongs to.
	* @param name the name of the method.
	* @param sig the signature of the method.
	* @return a MethodGen, or null if it could not be found. */
	public MethodGen getMethod(String className, String name, String sig) {
		ClassGen cgen = getClassGen(className);
		Method[] methods = cgen.getMethods();
		
		for(int i = 0; i < methods.length; i++) {
			Method m = methods[i];
			
			if(m.getName().equals(name) && m.getSignature().equals(sig)) {
				return new MethodGen(m, cgen.getClassName(), cgen.getConstantPool());
			}
		}
		return null;
	}

	/** gets a FieldGen object from a FieldID.
	* @param fid a FieldID.
	* @return a FieldGen, or null if it could not be found. */
	public FieldGen getField(FieldID fid) {
		return getField(fid.getClassName(), fid.getName(), fid.getSignature());
	}

	/** gets a FieldGen object from a FieldID.
	* @param className the class the field belongs to.
	* @param name the name of the field.
	* @param sig the signature of the field.
	* @return a FieldGen, or null if it could not be found. */
	public FieldGen getField(String className, String name, String sig) {
		ClassGen cgen = getClassGen(className);
		Field[] fields = cgen.getFields();
		
		for(int i = 0; i < fields.length; i++) {
			Field f = fields[i];
			
			if(f.getName().equals(name) && f.getSignature().equals(sig)) {
				return new FieldGen(f, cgen.getClassName(), cgen.getConstantPool());
			}
		}
		return null;
	}

	/** gets the names of all the classes held in this Application.
	* @return a Set of String class names. */
	public Set<String> getClassNames() {
		return classMap.keySet();
	}

	/** gets all the classes held in this Application.
	* @return a Collection of ClassGen objects. */
	public Collection<ClassGen> getClasses() {
		return classMap.values();
	}

	/** gets all the methods of every class in this Application.
	* @return an array of MethodGen objects. */
	public MethodGen[] allMethods() {
		ArrayList<MethodGen> list = new ArrayList<MethodGen>();
		
		for(ClassGen cgen : classMap.values()) {
			String className = cgen.getClassName();
			ConstantPoolGen cpg = cgen.getConstantPool();
			Method[] methods = cgen.getMethods();
			
			for(int i = 0; i < methods.length; i++) {
				list.add(new MethodGen(methods[i], className, cpg));
			}
		}
		return list.toArray(new MethodGen[0]);
	}

	/** gets all the fields of every class in this Application.
	* @return an array of FieldGen objects. */
	public FieldGen[] allFields() {
		ArrayList<FieldGen> list = new ArrayList<FieldGen>();
		
		for(ClassGen cgen : classMap.values()) {
			String className = cgen.getClassName();
			ConstantPoolGen cpg = cgen.getConstantPool();
			Field[] fields = cgen.getFields();
			
			for(int i = 0; i < fields.length; i++) {
				list.add(new FieldGen(fields[i], className, cpg));
			}
		}
		return list.toArray(new FieldGen[0]);
	}
}